#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Sep 19 20:46:21 2022

@author: mehlmann
"""

import numpy as np
import matplotlib.pyplot as plt

# we solve the model problem
#
#  u' = lambda u, u(0) = 1
#
# lambda is a complex number and for z = h * lambda we determine
# if the methods are stable by analyzing one single step
# if |u_1|>1 the method is unstable, otherwise it is stable

## One step of Runge-Kutta 4
def rk4(x,z):  # z = h * l
    k1 = z*x
    k2 = z*(x+0.5*k1)
    k3 = z*(x+0.5*k2)
    k4 = z*(x+k3)
    return x+1./6.*(k1+2.*k2+2.*k3+k4)

## One step of E Euler
def eeuler(x,z):  # z = h * l
    return x+z*x

## One step of Heun's method
def heun(x,z):  # z = h * l
    return x+z*x+0.5*z*z*x


## One step of Implicit Euler
def ieuler(x,z):  # z = h * l
    return x/(1.0-z*x)

## One step of Trapezoidal
def trapezoidal(x,z):  # z = h * l
    return (x+0.5*z*x)/(1.0-0.5*z*x)


## We create a mesh of points x/y from -6 to 6, real- and imaginary part
N = 200
lx = np.linspace(-4,4,N)
X,Y = np.meshgrid(lx,lx)

EE = np.zeros( (N,N) )  # explizit Euler
IE = np.zeros( (N,N) )  # implicit Euler
HE = np.zeros( (N,N) )  # Heun
RK = np.zeros( (N,N) )  # Runge Kutta
TR = np.zeros( (N,N) )  # Trapecoidal

for x in range(N): # real part
    for y in range(N): # imag part
        z = lx[x]+lx[y]*1.j            # 1.j is the imaginary unit
        
        if np.absolute(eeuler(1.,z))<=1:
            EE[y,x]=1
        if np.absolute(heun(1.,z))<=1:
            HE[y,x]=1
        if np.absolute(rk4(1.,z))<=1:
            RK[y,x]=1
        if np.absolute(ieuler(1.,z))<=1:
            IE[y,x]=1
        if np.absolute(trapezoidal(1.,z))<=1:
            TR[y,x]=1
 

plt.scatter(X,Y,EE,label='forward Euler')
plt.plot([0,0],[-4,4],'gray')
plt.plot([-4,4],[0,0],'gray')
plt.legend()
plt.show()


plt.scatter(X,Y,HE,label='Heun')
plt.plot([0,0],[-4,4],'gray')
plt.plot([-4,4],[0,0],'gray')
plt.legend()
plt.show()

plt.scatter(X,Y,RK,label='Runge-Kutta')
plt.plot([0,0],[-4,4],'gray')
plt.plot([-4,4],[0,0],'gray')
plt.legend()
plt.show()

plt.scatter(X,Y,IE,label='backward Euler')
plt.plot([0,0],[-4,4],'gray')
plt.plot([-4,4],[0,0],'gray')
plt.legend()
plt.show()

plt.scatter(X,Y,TR,label='Trapezoidal')
plt.plot([0,0],[-4,4],'gray')
plt.plot([-4,4],[0,0],'gray')
plt.legend()
plt.show()

