#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Sep 19 20:46:21 2022

@author: mehlmann
"""

import numpy as np
import matplotlib.pyplot as plt


# The matrix of the equation
# x'(t) = A x(t)
def A():
    return np.array([[0,1,0],
                     [-1,-0.01,-100],
                     [0,1,-1000]])


## One step of forward Euler
## [I - h A] u_n = u_{n-1}
def ieuler(x,h):
    
    # np.eye(3) is the identity matrix I
    # np.linalg.inv(np.eye(3)-h*A()) is the inverse of [I-hA]
    # np.dot is the matrix-vector product
    return np.dot(np.linalg.inv(np.eye(3)-h*A()),x)



## One step of trapecidal rule
## [I - 0.5 * h A] u_n = u_{n-1} + 0.5 * h A u_{n-1}
def trapecoidal(x,h):
    return np.dot(np.linalg.inv(np.eye(3)-0.5*h*A()),x+0.5*h*np.dot(A(),x))


## One step of Runge-Kutta 4
def rk4(x,h):
    k1 = np.dot(A(),x)
    k2 = np.dot(A(),x+0.5*h*k1)
    k3 = np.dot(A(),x+0.5*h*k2)
    k4 = np.dot(A(),x+h*k3)
    return x+h/6.*(k1+2.*k2+2.*k3+k4)


print('3. Implicit discretizations of the linear equation.\n')

print('a) backward Euler solution on [0,T]')
## Solve the problem in [0,10] and plot the solution
T = 10   # Final time
N = 10  # Number of time steps
h = T/N  # step size

timesteps = np.linspace(0,T,N+1)  # discrete time points

x = np.zeros( (N+1,3) )           # solution vector
x[0,:] = np.array([1,1,0])        # initial solution

for i in range(N):
    x[i+1,:] = ieuler(x[i,:],h)
plt.title('Exercise 3.a) Implicit Euler solution with h={0}'.format(h))
plt.plot(timesteps,x[:,0],label='x(t)')
plt.plot(timesteps,x[:,1],label="x'(t)")
plt.plot(timesteps,x[:,2],label='w(t)')
plt.legend()
plt.show()


print('\n3.b) and 3.d) We determine the order of convergence (as in Ex. 1)')

print('First we compute an exact solution with rk4 using h=0.001')
exact = np.array([1,1,0])  # set initial value
href = 0.001
for n in range(int(1.e-8+T/href)):
    exact = rk4(exact,href)

print('Reference solution is ',exact)
print('\n')

print('For h=1, 0.5, 0.25 we solve the problem and compute the error (with the reference solution)')
print('To measure the convergence rate we compute |xref - x_{h/2}| / |xref - x_h|')
print('Error reduction of 2 is linear convergence, 4 is quadratic, 16 is order 4')


M = 8  # number of refinements
errors = np.zeros((M,2))   #  for storing the errors


for hi in range(M):     # three successive refinements of the mehs size
    h = 0.5**hi   # h=0.002, h=0.001, h=0.0005
    
    N = int(1.e-8+T/h)  # set number of steps
    
    
    x = np.array([1,1,0])  # set initial value
    for i in range(N):
        x = ieuler(x,h)
    errors[hi,0] = np.linalg.norm(x-exact)
    
    x = np.array([1,1,0])  # set initial value
    for i in range(N):
        x = trapecoidal(x,h)
    errors[hi,1] = np.linalg.norm(x-exact)


## compute reduction factor
# error(h) / error(h/2)
reduction = errors.copy()
for m in reversed(range(M-1)):
    reduction[m+1,:] = errors[m,:]/errors[m+1,:]  
reduction[0,:] = np.zeros(2)

print('Errors and error reduction:\n')
print('h\t\tEuler\t\t\t\tTrapecoidal')

for hi in range(M):
    print('{0:2.6f}'.format(0.002*0.5**hi),end='\t')
    
    print('{0:2.4e} ({1:2.2f})\t{2:2.4e} ({3:2.2f})'.format
          (errors[hi,0],reduction[hi,0],errors[hi,1],reduction[hi,1]))

print('Euler is first order')
print('Trapecoidal is second order')



print('')
print('3.c) Comparison of Euler and Trapecoidal rule')

for h in [1., 0.5, 0.25]:
        
    T = 10              # Final time
    N = int(1.e-8+T/h)  # Number of steps

    timesteps = np.linspace(0,T,N+1)  # discrete time points

    xe = np.zeros( (N+1,3) )           # solution vector Euler
    xe[0,:] = np.array([1,1,0])        # initial solution
    xt = np.zeros( (N+1,3) )           # solution vector trapecoidal
    xt[0,:] = np.array([1,1,0])        # initial solution
    
    for i in range(N):  
        xe[i+1,:] = ieuler(xe[i,:],h) 
        xt[i+1,:] = trapecoidal(xt[i,:],h) 

    plt.title('3.c) Comparison of Euler and Trapecoidal')
    plt.plot(timesteps,xe[:,0],label='impl. Euler, h={0}'.format(h))
    plt.plot(timesteps,xt[:,0],label='trapecoidal rule, h={0}'.format(h))
    
    plt.legend()
    plt.show()