#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Sep 19 20:46:21 2022

@author: mehlmann
"""

import numpy as np
import matplotlib.pyplot as plt

# Implementation of the right hand side 
# x'(t) = f( x(t) )
def f(x):
    return np.array([x[1],
                     -x[0]-0.01*x[1]-100*x[1]*x[1]*x[2],
                     np.exp(2*x[1])-1000.0*x[2]])


## Implement the Jacobian Df(x)
def Df(x):
    return np.array([[0,1,0],
                     [-1, -0.01-200.0*x[1]*x[2],-100*x[1]*x[1]],
                     [0,np.exp(2*x[1])*2.,-1000]])


## One step of Implicity Euler with Newton
# x - h f(x) = o
# Newton:
#    [I - h Df(x)] w = o - (x - h f(x)) 
#    x = x + w 
def ieuler(x,h):
    xold = x
    for i in range(5): # We do 5 steps of newton
        w = np.dot(np.linalg.inv(np.eye(3)-h*Df(x)),xold - (x-h*f(x)))
        x = x + w
    return x

## One step of implicit trapecoidal rule with Newton
# x - 0.5 * h f(x) = o + 0.5 * h * f(o)
# Newton:
#    [I - 0.5 h Df(x)] w = o + 0.5 h f(0) - (x - 0.5 * h f(x)) 
#    x = x + w 
def trapecoidal(x,h):
    xold = x
    for i in range(5):  # 5 Newton iterations
        w = np.dot(np.linalg.inv(np.eye(3)-0.5*h*Df(x)),xold + 0.5*h*f(xold) - (x-0.5*h*f(x)))
        x = x + w
    return x

## One step of Heun's 2nd order method
def heun(x,h):
    return x+0.5*h*f(x) + 0.5*h*f(x+h*f(x))



## One step of Runge-Kutta 4
def rk4(x,h):
    k1 = f(x)
    k2 = f(x+0.5*h*k1)
    k3 = f(x+0.5*h*k2)
    k4 = f(x+h*k3)
    return x+h/6.*(k1+2.*k2+2.*k3+k4)

print('4. Nonlinear pendulum')
print('a) We use Heun''s method and plot the solution')

## Solve the problem in [0,10] and plot the solution
T = 10   # Final time
N = 6000  # Number of time steps
h = T/N  # step size

timesteps = np.linspace(0,T,N+1)  # discrete time points

x = np.zeros( (N+1,3) )           # solution vector
x[0,:] = np.array([1,1,0])        # initial solution

for i in range(N):
    x[i+1,:] = heun(x[i,:],h)

plt.title('4.a) Nonlinear spring pendulum')
plt.plot(timesteps,x,label='Heun''s solution')
plt.legend()
plt.show()


print('\n4.b) We solve with the implicit Euler and the Trapecoidal rule using h=0.1')

# Solve with IE
N = 50
h = T/N
timesteps = np.linspace(0,T,N+1)
x = np.zeros( (N+1,3) )           # solution vector
x[0,:] = np.array([1,1,0])        # initial solution

for i in range(N):
    x[i+1,:] = ieuler(x[i,:],h)
    
    
plt.title('Nonlinear pendulum, implicit methods h={0}'.format(h))
plt.plot(timesteps,x,label='Backward Euler')


# Solve with Trapecoidal
N = 50
h = T/N
timesteps = np.linspace(0,T,N+1)
x = np.zeros( (N+1,3) )           # solution vector
x[0,:] = np.array([1,1,0])        # initial solution

for i in range(N):
    x[i+1,:] = trapecoidal(x[i,:],h)
    
plt.plot(timesteps,x,label='Trapecoidal')
plt.legend()
plt.show()



print('\nc) We determine the order of convergence.')
print('First we compute an exact solution with rk4 using h=0.001')
exact = np.array([1,1,0])  # set initial value
href = 0.001
for n in range(int(1.e-8+T/href)):
    exact = rk4(exact,href)


## 

M = 8 # number of refinements
errors = np.zeros((M,2))   #  for storing the errors


for hi in range(M):     # three successive refinements of the mehs size
    h = 0.5**hi   # h=0.002, h=0.001, h=0.0005
    
    N = int(1.e-8+T/h)  # set number of steps
    
    
    x = np.array([1,1,0])  # set initial value
    for i in range(N):
        x = ieuler(x,h)
    errors[hi,0] = np.linalg.norm(x-exact)
    
    x = np.array([1,1,0])  # set initial value
    for i in range(N):
        x = trapecoidal(x,h)
    errors[hi,1] = np.linalg.norm(x-exact)


## compute reduction factor
# error(h) / error(h/2)
reduction = errors.copy()
for m in reversed(range(M-1)):
    reduction[m+1,:] = errors[m,:]/errors[m+1,:]  
reduction[0,:] = np.zeros(2)



print('Errors and error reduction:\n')
print('h\t\tEuler\t\t\t\tTrapecoidal')

for hi in range(M):
    print('{0:2.6f}'.format(0.002*0.5**hi),end='\t')
    
    print('{0:2.4e} ({1:2.2f})\t{2:2.4e} ({3:2.2f})'.format
          (errors[hi,0],reduction[hi,0],errors[hi,1],reduction[hi,1]))

